#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "ks0108.h"

struct displayPos pos;
uint8_t inverted=0;

void ks0108DrawCircle(uint8_t xCenter, uint8_t yCenter, uint8_t radius) {
  	int16_t tSwitch, y, x = 0;
  	uint8_t d;

  	d = yCenter - xCenter;
  	y = radius;
  	tSwitch = 3 - 2 * radius;
  	
  	while (x <= y) {
    	ks0108SetDot(xCenter + x, yCenter + y);
    	ks0108SetDot(xCenter + x, yCenter - y);
    	
    	ks0108SetDot(xCenter - x, yCenter + y);
    	ks0108SetDot(xCenter - x, yCenter - y);
    	
    	ks0108SetDot(yCenter + y - d, yCenter + x);
    	ks0108SetDot(yCenter + y - d, yCenter - x);
    	
    	ks0108SetDot(yCenter - y - d, yCenter + x);
    	ks0108SetDot(yCenter - y - d, yCenter - x);

    	if (tSwitch < 0) {
    		tSwitch += (4 * x + 6);
    	} else {
      		tSwitch += (4 * (x - y) + 10);
      		y--;
    	}
    	x++;
  	}
}

void ks0108DrawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2) {
	uint8_t length, xTmp, yTmp, i, y, yAlt;
	int16_t m;
	
	if(x1 == x2) {									// vertical line
		// x1|y1 must be the upper point
		if(y1 > y2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		length = y2-y1;
		for(i=0; i<=length; i++) {
			ks0108SetDot(x1, y1+i);
		}
	} else if(y1 == y2) {							// horizontal line
		// x1|y1 must be the left point
		if(x1 > x2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		length = x2-x1;
		for(i=0; i<=length; i++) {
			ks0108SetDot(x1+i, y1);
		}
	} else {
		// x1 must be smaller than x2
		if(x1 > x2) {
			xTmp = x1;
			yTmp = y1;
			x1 = x2;
			y1 = y2;
			x2 = xTmp;
			y2 = yTmp;
		}
		
		if((y2-y1) >= (x2-x1) || (y1-y2) >= (x2-x1)) {	// angle larger or equal 45
			length = x2-x1;								// not really the length :)
			m = ((y2-y1)*200)/length;
			yAlt = y1;
			for(i=0; i<=length; i++) {
				y = ((m*i)/200)+y1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				ks0108DrawLine(x1+i, yAlt, x1+i, y);
				if(length<=(y2-y1) && y1 < y2)
					yAlt = y+1;
				else if(length<=(y1-y2) && y1 > y2)
					yAlt = y-1;
				else
					yAlt = y;
			}
		} else {										// angle smaller 45
			// y1 must be smaller than y2
			if(y1 > y2) {
				xTmp = x1;
				yTmp = y1;
				x1 = x2;
				y1 = y2;
				x2 = xTmp;
				y2 = yTmp;
			}
			length = y2-y1;
			m = ((x2-x1)*200)/length;
			yAlt = x1;
			for(i=0; i<=length; i++) {
				y = ((m*i)/200)+x1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				ks0108DrawLine(yAlt, y1+i, y, y1+i);
				if(length<=(x2-x1) && x1 < x2)
					yAlt = y+1;
				else if(length<=(x1-x2) && x1 > x2)
					yAlt = y-1;
				else
					yAlt = y;
			}
		}
	}
}

void ks0108DrawRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height) {
	width--;
	height--;
	ks0108DrawLine(x, y, x+width, y);				// top
	ks0108DrawLine(x, y, x, y+height);				// left
	ks0108DrawLine(x, y+height, x+width, y+height);	// bottom
	ks0108DrawLine(x+width, y, x+width, y+height);	// right
}

void ks0108Enable(void) {
	volatile uint8_t i;
	
	LCD_CMD_PORT |= 0x01 << EN;						// EN high level width: min. 450ns
	asm volatile("nop\n\t"
				 "nop\n\t"
				 "nop\n\t"
				 ::);
	LCD_CMD_PORT &= ~(0x01 << EN);
	for(i=0; i<8; i++);								// a little delay loop (faster than reading the busy flag)
}

void ks0108Fill(uint8_t mode) {
	uint8_t i, j;
	
	for(i=0; i<8; i++) {							// page address
		ks0108GotoXY(0, i*8);
		for(j=0; j<128; j++)						// x address
			ks0108WriteData(mode);					// write data
	}
}

void ks0108GotoXY(uint8_t x, uint8_t y) {
	uint8_t chip = CHIP1, cmd;
	
	if(x > 127) x = 0;								// ensure that coordinates are legal
	if(y > 63)  y = 0;
	
	pos.x = x;										// save new coordinates
	pos.y = y;
	pos.page = y/8;
	
	if(x >= 64) {									// select the right chip
		x -= 64;
		chip = CHIP2;
	}
	cmd = LCD_SET_ADD | x;
	ks0108WriteCommand(cmd, chip);					// set x address on active chip
	
	cmd = LCD_SET_PAGE | pos.page;					// set y address on both chips
	ks0108WriteCommand(cmd, CHIP1);
	ks0108WriteCommand(cmd, CHIP2);
}

void ks0108Init(void) {
	pos.x = 0;
	pos.y = 0;
	pos.page = 0;
	
	LCD_CMD_DIR = 0xFF;								// command port is output
	LCD_CMD_PORT |= (1<<PC2);
	ks0108WriteCommand(LCD_ON, CHIP1);				// power on
	ks0108WriteCommand(LCD_ON, CHIP2);
	
	ks0108WriteCommand(LCD_DISP_START, CHIP1);		// display start line = 0
	ks0108WriteCommand(LCD_DISP_START, CHIP2);
	ks0108Fill(CLEAR);								// display clear
	ks0108GotoXY(0,0);
}

void ks0108InvertRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height) {
	uint8_t mask, pageOffset, pixels, i, data, tmpData;
	
	pageOffset = y%8;
	mask = 0xFF;
	if(height < 8-pageOffset) {
		mask >>= (8-height);
		pixels = height;
	} else {
		pixels = 8-pageOffset;
	}
	mask <<= pageOffset;
	
	ks0108GotoXY(x, y-pageOffset);
	for(i=0; i<width; i++) {
		data = ks0108ReadData(NO_INCREMENT_X);		// dummy read
		data = ks0108ReadData(NO_INCREMENT_X);		// "real" read
		tmpData = ~data;
		data = (tmpData & mask) | (data & ~mask);
		ks0108WriteData(data);
	}
	while(pixels+8 <= height) {
		ks0108GotoXY(x, pos.y+8);
		pixels += 8;
		for(i=0; i<width; i++) {
			data = ks0108ReadData(NO_INCREMENT_X);	// dummy read
			data = ks0108ReadData(NO_INCREMENT_X);	// "real" read
			ks0108WriteData(~data);
		}
	}
	if(pixels < height) {
		mask = ~(0xFF << (height-pixels));
		ks0108GotoXY(x, pos.y+8);
		for(i=0; i<width; i++) {
			data = ks0108ReadData(NO_INCREMENT_X);	// dummy read
			data = ks0108ReadData(NO_INCREMENT_X);	// "real" read
			tmpData = ~data;
			data = (tmpData & mask) | (data & ~mask);
			ks0108WriteData(data);
		}
	}
}

void ks0108SetDot(uint8_t x, uint8_t y) {
	uint8_t data;
	
	ks0108GotoXY(x, y-y%8);							// read data from display memory
	data = ks0108ReadData(NO_INCREMENT_X);			// dummy read
	data = ks0108ReadData(NO_INCREMENT_X);			// "real" read
	
	data |= 0x01 << (y%8);							// set dot
	
	ks0108WriteData(data);							// write data back to display
}

void ks0108ClearDot(uint8_t x, uint8_t y) {
	uint8_t data;
	
	ks0108GotoXY(x, y-y%8);							// read data from display memory
	data = ks0108ReadData(NO_INCREMENT_X);			// dummy read
	data = ks0108ReadData(NO_INCREMENT_X);			// "real" read
	
	data &= ~(0x01 << (y%8));						// clear dot
	
	ks0108WriteData(data);							// write data back to display
}

void ks0108PutChar(char c, struct font font) {
	uint16_t index;
	uint8_t pages, tmp, i, j, xPos, yPos;
	
	if(c == '\n')
		ks0108NewLine(font.height, 0);
	if(c < 32)										// ignore escape characters
		return;
	
	xPos = pos.x;									// save old coordinates
	yPos = pos.y;
	
	c -= 32;
	pages = font.height/8;							// calculate pages
	if(font.height%8 != 0)
		pages++;
	
	index = c*font.width*pages;						// get the needed array index

	for(j=0; j<pages; j++) {
		for(i=j; i<font.width*pages; i+=pages) {
			tmp = pgm_read_byte(font.charData+index+i);
			ks0108WriteData(tmp);					// write Character-Data
		}
		ks0108GotoXY(xPos, pos.y+8);
	}
	ks0108GotoXY(pos.x+font.width, yPos);			// go to the upper right corner
}

void ks0108NewLine(uint8_t fontHeight, uint8_t offset) {
	if(pos.y+fontHeight < 64)
		ks0108GotoXY(offset, pos.y+fontHeight);
	else
		ks0108GotoXY(offset, 0);
}

void ks0108PutString(char *string, struct font font) {
	uint8_t startx=pos.x, i=0;
	char c = string[0];
	
	while(c != 0) {
		if(c == '\n')
			ks0108NewLine(font.height, startx);
		else
			ks0108PutChar(c, font);
		c = string[++i];
	}
}

void ks0108PutStringP(PGM_P string, struct font font) {
	uint8_t startx=pos.x;
	char c = pgm_read_byte(string);
	
	while(c != 0) {
		if(c == '\n')
			ks0108NewLine(font.height, startx);
		else
			ks0108PutChar(c, font);
		
		c = pgm_read_byte(++string);
	}
}

void ks0108SetInverted(uint8_t invert) {
	inverted = invert;
}

char ks0108ReadData(uint8_t incXAdd) {
	uint8_t data;
	volatile uint8_t i;
	
	LCD_DATA_OUT = 0x00;
	LCD_DATA_DIR = 0x00;							// data port is input
	
	if(pos.x < 64) {
		LCD_CMD_PORT &= ~(0x01 << CSEL2);			// deselect chip 2
		LCD_CMD_PORT |= 0x01 << CSEL1;				// select chip 1
	} else if(pos.x >= 64) {
		LCD_CMD_PORT &= ~(0x01 << CSEL1);			// deselect chip 1
		LCD_CMD_PORT |= 0x01 << CSEL2;				// select chip 2
	}
	if(pos.x == 64)									// chip2 X-address = 0
		ks0108WriteCommand(LCD_SET_ADD, CHIP2);
	
	LCD_CMD_PORT |= 0x01 << D_I;					// D/I = 1
	LCD_CMD_PORT |= 0x01 << R_W;					// R/W = 1
	
	
	LCD_CMD_PORT |= 0x01 << EN;						// EN high level width: min. 450ns
	asm volatile("nop\n\t"
				 "nop\n\t"
				 "nop\n\t"
				 ::);
	
	data = LCD_DATA_IN;								// read Data			 
	
	LCD_CMD_PORT &= ~(0x01 << EN);
	for(i=0; i<8; i++);								// a little delay loop (faster than reading the busy flag)
	
	LCD_DATA_DIR = 0xFF;
	
	
	if(incXAdd == INCREMENT_X)						// display x-address is incremented automatically
		pos.x++;
	else
		ks0108GotoXY(pos.x, pos.y);
	
	if(inverted)
		data = ~data;
	return data;
}

void ks0108WriteCommand(uint8_t cmd, uint8_t chip) {
	if(chip == CHIP1) {
		LCD_CMD_PORT &= ~(0x01 << CSEL2);			// deselect chip 2
		LCD_CMD_PORT |= 0x01 << CSEL1;				// select chip 1
	} else if(chip == CHIP2) {
		LCD_CMD_PORT &= ~(0x01 << CSEL1);			// deselect chip 1
		LCD_CMD_PORT |= 0x01 << CSEL2;				// select chip 2
	}
	
	LCD_CMD_PORT &= ~(0x01 << D_I);					// D/I = 0
	LCD_CMD_PORT &= ~(0x01 << R_W);					// R/W = 0
	LCD_DATA_DIR = 0xFF;							// data port is output
	LCD_DATA_OUT = cmd;								// write command
	ks0108Enable();									// enable
	LCD_DATA_OUT = 0x00;
}

void ks0108WriteData(uint8_t data) {
	uint8_t displayData, yOffset, cmdPort;

#ifdef DEBUG
	volatile uint16_t i;
	for(i=0; i<5000; i++);
#endif

	if(pos.x >= 128)
		return;

	if(pos.x < 64) {
		LCD_CMD_PORT &= ~(0x01 << CSEL2);			// deselect chip 2
		LCD_CMD_PORT |= 0x01 << CSEL1;				// select chip 1
	} else if(pos.x >= 64) {
		LCD_CMD_PORT &= ~(0x01 << CSEL1);			// deselect chip 1
		LCD_CMD_PORT |= 0x01 << CSEL2;				// select chip 2
	}
	if(pos.x == 64)									// chip2 X-address = 0
		ks0108WriteCommand(LCD_SET_ADD, CHIP2);
	
	LCD_CMD_PORT |= 0x01 << D_I;					// D/I = 1
	LCD_CMD_PORT &= ~(0x01 << R_W);					// R/W = 0
	LCD_DATA_DIR = 0xFF;							// data port is output
	
	
	yOffset = pos.y%8;
	if(yOffset != 0) {
		// first page
		cmdPort = LCD_CMD_PORT;						// save command port
		displayData = ks0108ReadData(NO_INCREMENT_X);	// dummy read
		displayData = ks0108ReadData(NO_INCREMENT_X);	// "real" read
		
		LCD_CMD_PORT = cmdPort;						// restore command port
		LCD_DATA_DIR = 0xFF;						// data port is output
		
		displayData |= data << yOffset;
		if(inverted)
			displayData = ~displayData;
		LCD_DATA_OUT = displayData;					// write data
		ks0108Enable();								// enable
		
		// second page
		ks0108GotoXY(pos.x, pos.y+8);
		
		displayData = ks0108ReadData(NO_INCREMENT_X);	// dummy read
		displayData = ks0108ReadData(NO_INCREMENT_X);	// "real" read
		
		LCD_CMD_PORT = cmdPort;						// restore command port
		LCD_DATA_DIR = 0xFF;						// data port is output
		
		displayData |= data >> (8-yOffset);
		if(inverted)
			displayData = ~displayData;
		LCD_DATA_OUT = displayData;					// write data
		ks0108Enable();								// enable
		
		ks0108GotoXY(pos.x+1, pos.y-8);
	} else {
		if(inverted)
			data = ~data;
		LCD_DATA_OUT = data;						// write data
		ks0108Enable();								// enable
		pos.x++;
	}
	LCD_DATA_OUT = 0x00;
}
